@@ -9,6 +9,7 @@
 #include "../common/open.c"
 #include "../common/check.c"
 #include "../common/hash.c"
+#include "../common/mutex.c"
 #include "tap-interface.h"
 #include <stdlib.h>
 #include "logging.h"
@@ -22,12 +23,12 @@ static int tdb_expand_file_sparse(struct tdb_context *tdb,
 		return -1;
 	}
 
-	if (ftruncate(tdb->fd, size+addition) == -1) {
+	if (tdb_ftruncate(tdb, size+addition) == -1) {
 		char b = 0;
-		ssize_t written = pwrite(tdb->fd,  &b, 1, (size+addition) - 1);
+		ssize_t written = tdb_pwrite(tdb,  &b, 1, (size+addition) - 1);
 		if (written == 0) {
 			/* try once more, potentially revealing errno */
-			written = pwrite(tdb->fd,  &b, 1, (size+addition) - 1);
+			written = tdb_pwrite(tdb,  &b, 1, (size+addition) - 1);
 		}
 		if (written == 0) {
 			/* again - give up, guessing errno */
