@@ -116,23 +116,15 @@ static int epoll_wait_panic_fallback(struct epoll_event_context *epoll_ev,
 /*
   called to set the panic fallback function.
 */
-_PRIVATE_ bool tevent_epoll_set_panic_fallback(struct tevent_context *ev,
+_PRIVATE_ void tevent_epoll_set_panic_fallback(struct tevent_context *ev,
 				bool (*panic_fallback)(struct tevent_context *ev,
 						       bool replay))
 {
-	struct epoll_event_context *epoll_ev;
+	struct epoll_event_context *epoll_ev =
+		talloc_get_type_abort(ev->additional_data,
+		struct epoll_event_context);
 
-	if (ev->additional_data == NULL) {
-		return false;
-	}
-
-	epoll_ev = talloc_get_type(ev->additional_data,
-				struct epoll_event_context);
-	if (epoll_ev == NULL) {
-		return false;
-	}
 	epoll_ev->panic_fallback = panic_fallback;
-	return true;
 }
 
 /*
@@ -224,7 +216,7 @@ static void epoll_update_event(struct epoll_event_context *epoll_ev, struct teve
 
 /*
   reopen the epoll handle when our pid changes
-  see http://junkcode.samba.org/ftp/unpacked/junkcode/epoll_fork.c for an 
+  see http://junkcode.samba.org/ftp/unpacked/junkcode/epoll_fork.c for an
   demonstration of why this is needed
  */
 static void epoll_check_reopen(struct epoll_event_context *epoll_ev)
@@ -669,7 +661,7 @@ static int epoll_event_loop(struct epoll_event_context *epoll_ev, struct timeval
 	}
 
 	for (i=0;i<ret;i++) {
-		struct tevent_fd *fde = talloc_get_type(events[i].data.ptr, 
+		struct tevent_fd *fde = talloc_get_type(events[i].data.ptr,
 						       struct tevent_fd);
 		uint16_t flags = 0;
 		struct tevent_fd *mpx_fde = NULL;
@@ -843,8 +835,9 @@ static struct tevent_fd *epoll_event_add_fd(struct tevent_context *ev, TALLOC_CT
 					    const char *handler_name,
 					    const char *location)
 {
-	struct epoll_event_context *epoll_ev = talloc_get_type(ev->additional_data,
-							   struct epoll_event_context);
+	struct epoll_event_context *epoll_ev =
+		talloc_get_type_abort(ev->additional_data,
+		struct epoll_event_context);
 	struct tevent_fd *fde;
 	bool panic_triggered = false;
 
@@ -879,7 +872,8 @@ static void epoll_event_set_fd_flags(struct tevent_fd *fde, uint16_t flags)
 	if (fde->flags == flags) return;
 
 	ev = fde->event_ctx;
-	epoll_ev = talloc_get_type(ev->additional_data, struct epoll_event_context);
+	epoll_ev = talloc_get_type_abort(ev->additional_data,
+					 struct epoll_event_context);
 
 	fde->flags = flags;
 
@@ -894,12 +888,13 @@ static void epoll_event_set_fd_flags(struct tevent_fd *fde, uint16_t flags)
 }
 
 /*
-  do a single event loop using the events defined in ev 
+  do a single event loop using the events defined in ev
 */
 static int epoll_event_loop_once(struct tevent_context *ev, const char *location)
 {
-	struct epoll_event_context *epoll_ev = talloc_get_type(ev->additional_data,
-		 					   struct epoll_event_context);
+	struct epoll_event_context *epoll_ev =
+		talloc_get_type_abort(ev->additional_data,
+		struct epoll_event_context);
 	struct timeval tval;
 	bool panic_triggered = false;
 
@@ -908,6 +903,10 @@ static int epoll_event_loop_once(struct tevent_context *ev, const char *location
 		return 0;
 	}
 
+	if (ev->threaded_contexts != NULL) {
+		tevent_common_threaded_activate_immediate(ev);
+	}
+
 	if (ev->immediate_events &&
 	    tevent_common_loop_immediate(ev)) {
 		return 0;
